# rcsid = "$Id: circuit.tcl,v 1.1 2008/01/29 09:53:22 simon Exp $"

if [info exists env(CACD_SCRIPT_ROOT)] {
    source $env(CACD_SCRIPT_ROOT)/src/generic/libs/tcl/std/std.tcl
} else {
    source $env(CACD_DEVELOP_ROOT)/src/generic/libs/tcl/std/std.tcl
}

std::include_once src/generic/libs/tcl/std/options.tcl
std::include_once src/generic/libs/tcl/std/math.tcl
std::include_once src/generic/libs/tcl/std/path.tcl
std::include_once src/generic/libs/tcl/std/format.tcl
std::include_once src/generic/libs/tcl/std/exception.tcl
std::include_once src/generic/libs/tcl/std/array.tcl
std::include_once src/generic/libs/tcl/std/message.tcl
std::include_once src/generic/libs/tcl/std/validate.tcl

proc usage {} {
    set usage "Usage: circuit \[options\] \[--\] cell sub-command arg1 ... argN\n"
    append usage "Options:\n"
    append usage "  -h, --help                  Print this message.\n"
    append usage "  -v, --verbose               Turn on verbosity.\n"
    append usage "  --no-color                  Disable colored output.\n"
    append usage "\n"
    append usage "Examples:\n"
    append usage "    circuit -v oscil adjacent SUBSTR\n"
    append usage "    circuit -v oscil admittance oscil.admit\n"
    append usage "\n"
    return $usage
}

proc main {args} {

    set format_list [list]
    lappend format_list [list "--help"              0]
    lappend format_list [list "--verbose"           0]
    lappend format_list [list "--no-color"          0]

    lappend format_list [list "-h" "alias --help"]
    lappend format_list [list "-v" "alias --verbose"]

    array unset ::options

    std::try {
        std::options_parse $format_list $args ::options "auto_separator"
    } when user_exception {msg} {
        std::error_message $msg
        puts -nonewline stderr [usage]
        exit 1
    }

    if [info exists ::options(--no-color)] {
        std::disable_color
    }

    if [info exists ::options(--help)] {
        puts -nonewline stderr [usage]
        exit 0
    }

    if [info exists ::options(--verbose)] {
        std::enable_verbosity
    }

    set remain $::options(--)
    if {[llength $remain] < 2} {
        std::error_message "Please specify a cell name and a sub-command."
        puts -nonewline stderr [usage]
        exit 1
    }
    set cell_name [lindex $remain 0]
    set sub_command [lindex $remain 1]
    set sub_args [lrange $remain 2 end]

    std::top_eval {

        std::verbose "Loading cell `$cell_name' ..."
        set circuit [circuit load $cell_name]

        if {$sub_command == "adjacent"} {
            handle_adjacent $circuit $cell_name $sub_command $sub_args
        } elseif {$sub_command == "admittance"} {
            handle_admittance $circuit $cell_name $sub_command $sub_args
        } else {
            std::throw user_exception "Invalid sub-command `$sub_command'. Use the `-h' option for help."
        }

        std::verbose "Done"
    }
}

proc handle_adjacent {circuit cell_name sub_command sub_args} {

    set sub_format_list [list]
    lappend sub_format_list [list "--help" 0]
    lappend sub_format_list [list "--element-types"    "*"]
    lappend sub_format_list [list "--show-attributes"  0]

    lappend sub_format_list [list "-e" "alias --element-types"]
    lappend sub_format_list [list "-h" "alias --help"]
    lappend sub_format_list [list "-x" "alias --show-attributes"]

    array unset sub_options

    proc usage {} {
        set usage "Usage: circuit \"CELL\" adjacent \[options\] \[--\] \"NODE\"\n"
        append usage "Options:\n"
        append usage "  -h, --help\n"
        append usage "        Show this message.\n"
        append usage "\n"
        append usage "  -e, --element-types ELEM-TYPE ...\n"
        append usage "        Only consider the given element types.\n"
        append usage "        Valid element types are: \"res\" and \"cap\".\n"
        append usage "\n"
        append usage "  -x, --show-attributes\n"
        append usage "        Show also the attributes of each node.\n"
        append usage "\n"
        append usage "Example:\n"
        append usage "    circuit -v oscil adjacent --element-types res -- SUBSTR\n"
        append usage "\n"
        return $usage
    }

    std::try {
        std::options_parse $sub_format_list $sub_args sub_options
    } when user_exception {msg} {
        std::error_message $msg
        puts -nonewline stderr [usage]
        exit 1
    }

    if [info exists sub_options(--help)] {
        puts -nonewline stderr [usage]
        exit 0
    }

    set remain $sub_options(--)
    if {[llength $remain] != 1} {
        std::error_message "Specify a single node name after `adjacent'."
        puts -nonewline stderr [usage]
        exit 1
    }

    set nodes [$circuit info nodes]
    set node1 [lindex $remain 0]

    catch {unset models}
    if [info exists sub_options(--element-types)] {
        set models $sub_options(--element-types)
    }

    array unset table
    foreach elem [$circuit find elements $node1] {
        set model [$circuit element $elem model]
        if {[info exists models] && [lsearch -exact $models $model] < 0} {
            continue
        }
        
        foreach node [$circuit element $elem nodes] {
            if {$node == $node1} continue
            set table($node) 1
        }
    }

    foreach node [lsort -dictionary [array names table]] {
        if [info exists sub_options(--show-attributes)] {
            set attrs [$circuit node coordinates $node]
            puts "$node $attrs"
        } else {
            puts $node
        }
    }
}

proc handle_admittance {circuit cell_name sub_command sub_args} {

    set sub_format_list [list]
    lappend sub_format_list [list "--help" 0]
    lappend sub_format_list [list "-h" "alias --help"]

    array unset sub_options

    proc usage {} {
        set usage "Usage: circuit \"CELL\" admittance \[options\] \[--\] \"SCRIPT\"\n"
        append usage "Options:\n"
        append usage "  -h, --help                  Show this message.\n"
        append usage "\n"
        append usage "Example:\n"
        append usage "    circuit -v oscil admittance oscil.admit\n"
        append usage "\n"
        return $usage
    }

    std::try {
        std::options_parse $sub_format_list $sub_args sub_options
    } when user_exception {msg} {
        std::error_message $msg
        puts -nonewline stderr [usage]
        exit 1
    }

    if [info exists sub_options(--help)] {
        puts -nonewline stderr [usage]
        exit 0
    }

    set remain $sub_options(--)
    if {[llength $remain] != 1} {
        std::error_message "Specify a single filename after `admittance'."
        puts -nonewline stderr [usage]
        exit 1
    }
    set filename [lindex $remain 0]

    load_admittance_script $filename parse_table

    # Make sure we have a node called `0', which is grounded.
    # 
    catch {$circuit create node 0}

    # Some nodes which are implicitly grounded.
    #
    set parse_table(grounded,GND) 1
    set parse_table(grounded,gnd) 1

    foreach gnd [array names parse_table grounded,*] {
        regsub {^grounded,} $gnd "" gnd
        if {$gnd == "0"} {continue}
        catch {$circuit collapse "0 $gnd"}
    }

    # Create a very small artificial conductance to ground, from each node.
    #
    foreach node [$circuit info nodes] {
        if {$node == "0"} continue
        $circuit create element res "v=1e12" 0 $node
    }

    # Create a very small artificial capacitance to ground, from each node.
    #
    foreach node [$circuit info nodes] {
        if {$node == "0"} continue
        $circuit create element cap "v=1e-21" 0 $node
    }

    array unset node_table
    foreach node [$circuit info nodes] {
        set node_table($node) 1
    }

    # Compute two-port admittances.
    #
    foreach twoport $parse_table(twoport_list) {

        std::verbose "Twoport $twoport ..." 

        set node1 [lindex $twoport 0]
        set node2 [lindex $twoport 1]

        if ![info exists node_table($node1)] {
            std::throw user_exception "Invalid node `$node1'."
        }
        if ![info exists node_table($node2)] {
            std::throw user_exception "Invalid node `$node2'."
        }
        if [string equal $node1 $node2] {
            std::throw user_exception "Same node `$node1' used for both sides of twoport."
        }

        if {[lsearch -exact $parse_table(show) "spice"] >= 0} {
            set spice_filename "twoport_$twoport.spc"
            regsub -all " " $spice_filename "_" spice_filename

            std::verbose "Writing SPICE file `$spice_filename' ..."
            set fp [open $spice_filename w]
            puts -nonewline $fp [generate_admittance_spice_deck $circuit $node1 $node2]
            close $fp
        }

        foreach frequency $parse_table(frequencies) {
            if [catch {
                set r1 [$circuit compute_twoport_admittances 0 $node1 $node2 $frequency]
                set r2 [$circuit compute_twoport_admittances 0 $node2 $node1 $frequency] 
            } msg] {
                puts "twoport: $twoport: f=[format %.5g $frequency]: #numerical-error"
                continue
            }

            array unset table
            set table(y11) [lindex $r1 0]
            set table(y12) [lindex $r1 1]
            set table(y21) [lindex $r1 1]
            set table(y22) [lindex $r2 0]
            set table(y10) [std::Plus $table(y11) [std::Times {-1.0 0.0} $table(y12)]]
            set table(y20) [std::Plus $table(y22) [std::Times {-1.0 0.0} $table(y12)]]

            foreach ypar {y11 y12 y21 y22 y10 y20} {
                regsub y $ypar r rpar
                set table($rpar) "#unknown" 
                catch {set table($rpar) [std::Power $table($ypar) {-1.0 0.0}]}
            }

            foreach par $parse_table(show) {
                if ![info exists table($par)] continue
                set string "#unknown"
                catch {set string [std::format_complex $table($par)]}
                puts "twoport: $twoport: f=[format %.5g $frequency]: $par=$string"
            }
        }
    }
}


#===============================================================================
#
#  Load the "admittance" script.
#
#  The script should have the form:
#
#   cell: "cell-name"
#   grounded: "node1" "node2" "node3"
#   twoport: "node4" "node5"
#
#  Thus, one line per directive. Comments allowed using "#" mark.
#  There may be multiple "grounded" and "twoport" clauses.
#
#===============================================================================

proc load_admittance_script {filename _parse_table} {

    upvar $_parse_table parse_table
    array unset parse_table

    set parse_table(twoport_list) [list]
    set parse_table(cell_name) ""
    set parse_table(frequencies) [list 0.0]
    set parse_table(show) [list y11 y12]

    std::verbose "Loading script `$filename' ..."

    set fp [open $filename r]
    set lineno 0
    while {[gets $fp line] >= 0} {
        incr lineno
        regsub -all "#.*" $line "" line
        set list [list]
        foreach elem [split $line " :"] {
            if [string length $elem] {
                lappend list $elem
            }
        }
        if ![llength $list] continue
        set type [lindex $list 0]
        set list [lrange $list 1 end]
        set found 1
        std::try {
            if {$type == "grounded"} {
                foreach elem $list {
                    set parse_table(grounded,$elem) 1
                }
            } elseif {$type == "twoport"} {
                if {[llength $list] != 2} {
                    std::throw user_exception "Should specify two node names."
                }
                lappend parse_table(twoport_list) $list
            } elseif {$type == "sweep"} {
                if {[llength $list] != 3} {
                    std::throw user_exception "Should specify two frequencies and a number of steps."
                }
                set f1 [std::validate_double [lindex $list 0] 1]
                set f2 [std::validate_double [lindex $list 1] 1]
                set steps [std::validate_integer [lindex $list 2] 1]

                set parse_table(frequencies) [list]
                for {set s 0} {$s <= $steps} {incr s} {
                    set f [expr exp(log($f1) + $s*log($f2/$f1)/($steps))]
                    lappend parse_table(frequencies) $f
                }
            } elseif {$type == "show"} {
                set parse_table(show) $list
            } elseif {$type == "f" || $type == "frequencies"} {
                if {[llength $list] == 0} {
                    std::throw user_exception "Should specify at least one frequency."
                }

                set parse_table(frequencies) [list]
                foreach e $list {
                    set f [std::validate_double $e 1]
                    lappend parse_table(frequencies) $f
                }
            } else {
                set found 0
            }
        } when user_exception {msg} {
            std::throw user_exception "$filename: line $lineno: $msg"
        }

        if !$found {
            std::throw user_exception "$filename: line $lineno: Unknown directive `$type'."
        }
    }
    close $fp
}


#===============================================================================
#
#  Generate a SPICE deck for "admittance" computations.
#
#===============================================================================

proc generate_admittance_spice_deck {circuit node1 node2} {

    array unset element_table
    array unset node_table
    foreach node [$circuit info nodes] {
        set node_table($node) $node
        foreach element [$circuit find elements $node] {
            set element_table($element) 1
        }
    }

    set last_id 0
    foreach node [array names node_table] {
        if [regexp {^[0-9]+$} $node] continue

        while 1 {
            set ext_node [incr last_id]
            if ![info exists node_table($ext_node)] break
        }

        set node_table($node) $ext_node
    }

    set string ""
    append string "$circuit\n"
    append string "\n"

    append string "********************************************************************************\n"
    append string "* USAGE:\n"
    append string "*\n"
    append string "*   Simply run the following command from the unix-prompt:\n"
    append string "*\n"
    append string "*   # spice3 -b filename.spc\n"
    append string "*\n"
    append string "********************************************************************************\n"
    append string "* NODE MAPPING:\n"
    append string "*\n"
    foreach node [lsort -dictionary [array names node_table]] {
        append string "* [format %-10s $node] => [format %-10s $node_table($node)]\n"
    }
    append string "********************************************************************************\n"
    append string "\n"

    foreach element [array names element_table] {
        set model [$circuit element $element model]
        if {$model == "res"} {
            append string "r[incr last_id]"
        } elseif {$model == "cap"} {
            append string "c[incr last_id]"
        } else {
            error "Unknown circuit model `$model'."
        }

        foreach node [$circuit element $element nodes] {
            append string " $node_table($node)"
        }

        append string " [$circuit element $element attribute get v]"
        append string "\n"
    }
    append string "\n"

    append string "vin $node_table($node1) 0 0 ac 1\n"
    append string "vout $node_table($node2) 0 0\n"
    append string "\n"

    append string ".ac dec 100 1e1 1e20\n"
    append string ".plot ac abs(i(vin))\n"
    append string ".plot ac abs(i(vout))\n"
    append string ".end\n"

    return $string
}


#===============================================================================
#
#  Invoke main.
#
#===============================================================================

eval main $argv

