
################################################################################
namespace eval std {
################################################################################

proc date_split {date _year _month _day} {
    upvar $_year year
    upvar $_month month
    upvar $_day day
    std::invariant {[string length $date] == 8}
    set year [string range $date 0 3]
    regsub "^0*" $year "" year
    set year [expr int($year)]
    set month [string range $date 4 5]
    regsub "^0*" $month "" month
    set month [expr int($month)]
    set day [string range $date 6 7]
    regsub "^0*" $day "" day
    set day [expr int($day)]
}

proc date_get_month_size {year month} {
    set list {0 31 28 31 30 31 30 31 31 30 31 30 31}
    std::invariant {$month >= 1 && $month <= 12}
    set size [lindex $list $month]
    if {$month == 2 && [date_is_leap_year $year]} {
        incr size
    }
    return $size
}

proc date_get_year_size {year} {
    if [date_is_leap_year $year] {
        return 366
    } else {
        return 365
    }
}

proc date_is_leap_year {year} {
    if [expr !($year%400)] {return 1}
    if [expr !($year%100)] {return 0}
    if [expr !($year%4)] {return 1}
    return 0
}

proc date_is_valid {date} {
    date_split $date year month day
    if {$year < 0} return 0;
    if {$day < 1 || $day > [date_get_month_size $year $month]} return 0;
    if {$month < 1 || $month > 12} return 0
    return 1
}

proc get_current_date {} {
    set date [clock format [clock seconds] -format "%Y%m%d"]
}

proc date_to_epoch_time {date} {
    date_split $date year month day
    set e [expr $day-1]
    set arr {0 0 31 59 90 120 151 181 212 243 273 304 334 365}
    incr e [lindex $arr $month]
    if {[date_is_leap_year $year] && $month > 2} [incr e]
    incr e [expr $year*365]
    set ly [expr int(($year-1)/4)]
    set ly [expr $ly-int(($year-1)/100)]
    set ly [expr $ly+int(($year-1)/400)]
    if {$year != 0} {incr ly}
    incr e $ly
    return $e
}

proc date_from_epoch_time {e} {
    set y [expr int(double($e)/365.25)]
    set ly [expr int(($y-1)/4)]
    set ly [expr $ly-int(($y-1)/100)]
    set ly [expr $ly+int(($y-1)/400)]
    if {$y != 0} {incr ly}
    set eh [expr $y*365+$ly]
    set e [expr $e-$eh]

    set ys [date_get_year_size $y]
    if {$e >= $ys} { 
        incr y
        set e [expr $e-$ys]
    }
    set year $y
    set m [expr int(($e/31)+1)]

    set arr {0 0 31 59 90 120 151 181 212 243 273 304 334 365}
    set e [expr $e - [lindex $arr $m]]
    if {[date_is_leap_year $y] && $m > 2} {incr e -1}
    set day [expr $e+1]
    if {$day > [date_get_month_size $y $m]} {
        set day [expr $day-[date_get_month_size $y $m]]
        incr m
    }
    set month $m
    
    return [date_join $year $month $day]
}

proc date_join {year month day} {
    return [format %04d%02d%02d $year $month $day]
}


proc date_add_days {date n} {
    set e [date_to_epoch_time $date]
    incr e $n
    set date [date_from_epoch_time $e]
    return $date
}

proc date_add_one_day {date} {
    date_split $date year month day
    incr day 1
    if {$day > [date_get_month_size $year $month]} {
        incr month
        set day 1
        if {$month > 12} {
            incr year
            set month 1
        }
    }
    return [date_join $year $month $day]
}

proc date_add_months {date n} {
    date_split $date year month day
    set m [expr $month+$n-1]
    if {$m > 11} {
        incr year [expr int($m/12)]
        set m [expr int($m%12)]
    } elseif {$m < 0} {
        set year_min [expr int((-$m+11)/12)]
        incr m [expr $year_min*12]
        set year [expr $year - $year_min]
    }

    set month [expr $m+1]
    return [date_join $year $month $day]
}

proc date_add_years {date n} {
    date_split $date year month day
    incr year $n
    return [date_join $year $month $day]
}

proc date_diff_days {date1 date2} {
    set e1 [date_to_epoch_time $date1]
    set e2 [date_to_epoch_time $date2]
    return [expr $e1-$e2]
}

    # Day of week: 0=mon, 1=tue, etc.
    #
proc date_get_day_of_week {date} {
    set e [date_to_epoch_time $date]
    return [expr ($e+5)%7]
}


#===============================================================================
#!  Get formatted current date.
#===============================================================================

proc get_formatted_current_date {} {
    set date_string [clock format [clock seconds] -format "%a %b %d %H:%M:%S %Z %Y" -gmt 1]
    append date_string " GMT"
    return $date_string
}


################################################################################
}
################################################################################

