# rcsid = "$Id: layout.tcl,v 1.1 2008/01/29 09:10:32 simon Exp $"

if [info exists env(CACD_SCRIPT_ROOT)] {
    source $env(CACD_SCRIPT_ROOT)/src/generic/libs/tcl/std/std.tcl
} else {
    source $env(CACD_DEVELOP_ROOT)/src/generic/libs/tcl/std/std.tcl
}

std::include_once src/generic/libs/tcl/std/options.tcl
std::include_once src/generic/libs/tcl/std/classes.tcl
std::include_once src/generic/libs/tcl/std/path.tcl
std::include_once src/generic/libs/tcl/std/exec.tcl
std::include_once src/generic/libs/tcl/std/format.tcl
std::include_once src/generic/libs/tcl/std/validate.tcl
std::include_once src/generic/libs/tcl/std/message.tcl
std::include_once src/generic/libs/tcl/std/exception.tcl
std::include_once src/generic/libs/tcl/std/algorithm.tcl
std::include_once src/generic/libs/tcl/std/array.tcl
std::include_once src/generic/libs/tcl/std/constants.tcl
std::include_once src/generic/libs/tcl/std/math.tcl
std::include_once src/generic/libs/tcl/std/obj.tcl

std::include_once src/space/utilities/libs/tcl/project_tools/tools.tcl


#===============================================================================
#
#  Main.
#
#===============================================================================

proc main {args} {

    set format_list [list]
    lappend format_list [list "--help"          0]
    lappend format_list [list "--verbose"       0]
    lappend format_list [list "--no-color"      0]
    lappend format_list [list "--logical-width" 1]
    lappend format_list [list "--logical-height" 1]
    lappend format_list [list "--cell-size"     1]
    lappend format_list [list "--mask-sets"     *]

    lappend format_list [list "-h" "alias --help"]
    lappend format_list [list "-v" "alias --verbose"]
    lappend format_list [list "-W" "alias --logical-width"]
    lappend format_list [list "-H" "alias --logical-height"]
    lappend format_list [list "-C" "alias --cell-size"]
    lappend format_list [list "-M" "alias --mask-sets"]

    set usage ""
    append usage "Usage: layout \[options\] CELL-NAME\n"
    append usage "Options:\n"
    append usage "  -h, --help                  Print this message.\n"
    append usage "  -v, --verbose               Turn on verbosity.\n"
    append usage "  -W, --logical-width VALUE   Set the logical width of the layout.\n"
    append usage "                              (default 10).\n"
    append usage "  -H, --logical-height VALUE  Set the logical height of the layout.\n"
    append usage "                              (default 10).\n"
    append usage "  -C, --cell-size VALUE       Set the size of one logical cell\n"
    append usage "                              in microns (default 1.0).\n"
    append usage "  -M, --mask-sets SET1 ... SETn\n"
    append usage "                              Define the mask-sets.\n"
    append usage "  --no-color                  Disable colored output.\n"

    std::try {
        array unset ::main_options
        std::options_parse $format_list $args ::main_options
    } when user_exception {msg} {
        puts stderr $msg
        puts stderr $usage
        exit 1
    }

    if [info exists ::main_options(--no-color)] {
        std::disable_color
    }

    if [info exists ::main_options(--verbose)] {
        std::enable_verbosity
    }

    if [info exists ::main_options(--help)] {
        puts stderr $usage
        exit 0
    }

    array unset geometry
    set geometry(columns) 10
    set geometry(rows) 10

    if [info exists ::main_options(--logical-width)] {
        set geometry(columns) [std::validate_integer [lindex $::main_options(--logical-width) 0] 1]
    }
    if [info exists ::main_options(--logical-height)] {
        set geometry(rows) [std::validate_integer [lindex $::main_options(--logical-height) 0] 1]
    }

    if {[llength $::main_options(--)] != 1} {
        std::error_message "Please specify exactly one cell-name."
        puts stderr $usage
        exit 1
    }
    set cell_name [lindex $::main_options(--) 0]

    # Determine the project lambda.
    #
    set lambda [::project_tools::determine_lambda [pwd]]

    set cell_size 1.0
    if [info exists ::main_options(--cell-size)] {
        set cell_size [std::validate_double [lindex $::main_options(--cell-size) 0] 1]
    }
    set cell_units [expr int([std::rint [expr $cell_size/$lambda]])]
    if {$cell_units <= 1} {
        std::throw user_exception "Resolution too low. Please use a finer project lambda value."
    }
    set gap_units [expr int(double($cell_units)*0.3)]
    if {$gap_units == 0} {set gap_units 1}
    set square_units [expr $cell_units-$gap_units]

    std::verbose "Cell units: $cell_units"
    std::verbose "Gap units: $gap_units"
    std::verbose "Square units: $square_units"

    if [info exists ::main_options(--mask-sets)] {
        set mask_sets $::main_options(--mask-sets)
    } else {
        set mask_sets [list {cmf} {cms} {cpg}]
    }

    set fp [open "$cell_name.ldm" w]
    puts $fp ":: Generated by `layout' on [clock format [clock seconds]]."
    puts $fp "ms $cell_name"

    set mask_set_index -1
    foreach mask_set $mask_sets {

        incr mask_set_index

        array unset maze
        set maze(columns) $geometry(columns)
        set maze(rows) $geometry(rows)

        expr srand([clock clicks])
        generate_maze maze
        seek_path 0 0 [expr $maze(columns)-1] [expr $maze(rows)-1] maze

        if [std::is_verbose] {
            std::verbose "Layout for mask-set `$mask_set':"
            print_maze stderr maze
        }

        for {set y 0} {$y < $maze(rows)} {incr y} {
            for {set x 0} {$x < $maze(columns)} {incr x} {
                if $maze(path:c:$x:$y) {
                    foreach mask $mask_set {
                        puts $fp "box $mask [expr $cell_units*$x] [expr $cell_units*$x+$square_units] [expr $cell_units*$y] [expr $cell_units*$y+$square_units]"
                    }
                }
                if $maze(path:v:$x:$y) {
                    foreach mask $mask_set {
                        puts $fp "box $mask [expr $cell_units*$x] [expr $cell_units*$x+$square_units] [expr $cell_units*$y+$square_units] [expr $cell_units*$y+$square_units+$gap_units+1]"
                    }
                }
                if $maze(path:h:$x:$y) {
                    foreach mask $mask_set {
                        puts $fp "box $mask [expr $cell_units*$x+$square_units] [expr $cell_units*$x+$square_units+$gap_units+1] [expr $cell_units*$y] [expr $cell_units*$y+$square_units]"
                    }
                }
            }
        }

        set x 0
        set y 0
        set term_name "in_$mask_set_index"
        set index -1
        foreach mask $mask_set {
            incr index
            set real_term_name $term_name
            if {$index != 0} {set real_term_name "${term_name}_$index"}
            puts $fp "term $mask [expr $cell_units*$x] [expr $cell_units*$x+$square_units] [expr $cell_units*$y] [expr $cell_units*$y+$square_units] $real_term_name"
        }

        set x [expr $maze(columns)-1]
        set y [expr $maze(rows)-1]
        set term_name "out_$mask_set_index"
        set index -1
        foreach mask $mask_set {
            incr index
            set real_term_name $term_name
            if {$index != 0} {set real_term_name "${term_name}_$index"}
            puts $fp "term $mask [expr $cell_units*$x] [expr $cell_units*$x+$square_units] [expr $cell_units*$y] [expr $cell_units*$y+$square_units] $real_term_name"
        }
    }

    puts $fp "me"
    puts $fp ":: eof"
    close $fp

    std::quiet_execute cldm -f $cell_name.ldm

    std::verbose "Done."
    exit 0
}


#===============================================================================
#
#  Generate maze.
#
#===============================================================================

proc generate_maze {_maze} {

    upvar $_maze maze

    eval {
        set index 0
        for {set x 0} {$x < $maze(columns)} {incr x} {
            for {set y 0} {$y < $maze(rows)} {incr y} {
                set maze(h:$x:$y) 0
                set maze(v:$x:$y) 0
                set maze(c:$x:$y) 0

                set maze(path:h:$x:$y) 0
                set maze(path:v:$x:$y) 0
                set maze(path:c:$x:$y) 0
            }
        }
    }

    set x [expr int(rand()*$maze(columns))]
    set y [expr int(rand()*$maze(rows))]

    set maze(locations:size) 1
    set maze(locations:0:x) $x
    set maze(locations:0:y) $y
    set maze(c:$x:$y) 1

    while {$maze(locations:size) > 0} {
        set l [expr int(rand()*$maze(locations:size))]

        set x $maze(locations:$l:x)
        set y $maze(locations:$l:y)

        set found 0
        foreach dir [shuffle "h+ h- v+ v-"] {
            apply_direction $x $y $dir nx ny
            if {$nx < 0 || $ny < 0 || $nx >= $maze(columns) || $ny >= $maze(rows)} continue
            if $maze(c:$nx:$ny) continue

            set found 1

            if {$dir == "h+"} {
                set maze(h:$x:$y) 1
            } elseif {$dir == "h-"} {
                set maze(h:$nx:$ny) 1
            } elseif {$dir == "v+"} {
                set maze(v:$x:$y) 1
            } elseif {$dir == "v-"} {
                set maze(v:$nx:$ny) 1
            }
            add_location $nx $ny maze
            set maze(c:$nx:$ny) 1

            draw_path $nx $ny maze
        }

        if !$found {
            remove_location $l maze
        }
    }
}

proc add_location {x y _maze} {
    upvar $_maze maze
    set l $maze(locations:size)
    set maze(locations:$l:x) $x
    set maze(locations:$l:y) $y
    incr maze(locations:size)
    return $l
}

proc remove_location {l _maze} {
    upvar $_maze maze
    set from [expr $maze(locations:size)-1]
    set maze(locations:$l:x) $maze(locations:$from:x)
    set maze(locations:$l:y) $maze(locations:$from:y)
    incr maze(locations:size) -1
}

proc draw_path {x y _maze} {
    upvar $_maze maze

    while 1 {
        set found 0
        foreach dir [shuffle "h+ h- v+ v-"] {
            apply_direction $x $y $dir nx ny
            if {$nx < 0 || $ny < 0 || $nx >= $maze(columns) || $ny >= $maze(rows)} continue
            if $maze(c:$nx:$ny) continue

            if {$dir == "h+"} {
                set maze(h:$x:$y) 1
            } elseif {$dir == "h-"} {
                set maze(h:$nx:$ny) 1
            } elseif {$dir == "v+"} {
                set maze(v:$x:$y) 1
            } elseif {$dir == "v-"} {
                set maze(v:$nx:$ny) 1
            }
            add_location $nx $ny maze
            set maze(c:$nx:$ny) 1

            set found 1
            set x $nx
            set y $ny
            break
        }

        if !$found break
    }
}

proc apply_direction {x y dir _nx _ny} {
    upvar $_nx nx
    upvar $_ny ny

    if {$dir == "h+"} {
        set nx [expr $x+1]
        set ny $y
    } elseif {$dir == "h-"} {
        set nx [expr $x-1]
        set ny $y
    } elseif {$dir == "v+"} {
        set nx $x
        set ny [expr $y+1]
    } elseif {$dir == "v-"} {
        set nx $x
        set ny [expr $y-1]
    }
}

proc shuffle {list} {

    set new_list [list]
    while {[llength $list]} {
        set i [expr int(rand()*[llength $list])]
        lappend new_list [lindex $list $i]
        set list [lreplace $list $i $i]
    }
    return $new_list
}

proc print_maze {fp _maze} {
    upvar $_maze maze

    puts -nonewline $fp "*"
    for {set x 0} {$x < $maze(columns)} {incr x} {
        puts -nonewline $fp "**"
    }
    puts $fp ""

    for {set y 0} {$y < $maze(rows)} {incr y} {
        puts -nonewline $fp "*"
        for {set x 0} {$x < $maze(columns)} {incr x} {
            if $maze(c:$x:$y) {
                if $maze(path:c:$x:$y) {
                    puts -nonewline $fp "0"
                } else {
                    puts -nonewline $fp " "
                }
            } else {
                puts -nonewline $fp "#"
            }
            if $maze(h:$x:$y) {
                if $maze(path:h:$x:$y) {
                    puts -nonewline $fp "0"
                } else {
                    puts -nonewline $fp " "
                }
            } else {
                puts -nonewline $fp "*"
            }
        }
        puts $fp ""
        puts -nonewline $fp "*"
        for {set x 0} {$x < $maze(columns)} {incr x} {
            if $maze(v:$x:$y) {
                if $maze(path:v:$x:$y) {
                    puts -nonewline $fp "0"
                } else {
                    puts -nonewline $fp " "
                }
            } else {
                puts -nonewline $fp "*"
            }
            puts -nonewline $fp "*"
        }
        puts $fp ""
    }
}

proc seek_path {x y x2 y2 _maze} {
    upvar $_maze maze

    set maze(path:c:$x:$y) 1

    if {$x == $x2 && $y == $y2} {
        return 1
    }

    set found 0
    foreach dir "h+ h- v+ v-" {

        apply_direction $x $y $dir nx ny
        if {$nx < 0 || $ny < 0 || $nx >= $maze(columns) || $ny >= $maze(rows)} continue
        if $maze(path:c:$nx:$ny) continue

        if {$dir == "h+"} {
            if !$maze(h:$x:$y) continue
        } elseif {$dir == "h-"} {
            if !$maze(h:$nx:$ny) continue
        } elseif {$dir == "v+"} {
            if !$maze(v:$x:$y) continue
        } elseif {$dir == "v-"} {
            if !$maze(v:$nx:$ny) continue
        }

        set found [seek_path $nx $ny $x2 $y2 maze]
        if {$found} {
            if {$dir == "h+"} {
                set maze(path:h:$x:$y) 1
            } elseif {$dir == "h-"} {
                set maze(path:h:$nx:$ny) 1
            } elseif {$dir == "v+"} {
                set maze(path:v:$x:$y) 1
            } elseif {$dir == "v-"} {
                set maze(path:v:$nx:$ny) 1
            }

            break
        }
    }

    if !$found {
        set maze(path:c:$x:$y) 0
    }

    return $found
}


#===============================================================================
#
#  Invoke main.
#
#===============================================================================

std::top_eval {
    eval main $argv
}

