# rcsid = "$Id: tabs-verify.tcl,v 1.1 2008/01/29 09:05:57 simon Exp $"

if [info exists env(CACD_SCRIPT_ROOT)] {
    source $env(CACD_SCRIPT_ROOT)/src/generic/libs/tcl/std/std.tcl
} else {
    source $env(CACD_DEVELOP_ROOT)/src/generic/libs/tcl/std/std.tcl
}

std::include_once src/generic/libs/tcl/std/options.tcl
std::include_once src/generic/libs/tcl/std/classes.tcl
std::include_once src/generic/libs/tcl/std/path.tcl
std::include_once src/generic/libs/tcl/std/format.tcl
std::include_once src/generic/libs/tcl/std/validate.tcl
std::include_once src/generic/libs/tcl/std/message.tcl
std::include_once src/generic/libs/tcl/std/exception.tcl
std::include_once src/generic/libs/tcl/std/algorithm.tcl
std::include_once src/generic/libs/tcl/std/array.tcl
std::include_once src/generic/libs/tcl/std/constants.tcl
std::include_once src/generic/libs/tcl/std/math.tcl
std::include_once src/generic/libs/tcl/std/obj.tcl


#===============================================================================
#
#  Main.
#
#===============================================================================

proc main {args} {

    set format_list [list]
    lappend format_list [list "--help"      0]
    lappend format_list [list "--verbose"   0]
    lappend format_list [list "--verboser"  0]
    lappend format_list [list "--no-color"  0]
    lappend format_list [list "--precision" 1]
    lappend format_list [list "--pfile"     1]
    lappend format_list [list "--tfile"     1]
    lappend format_list [list "--min-fs"    1]
    lappend format_list [list "--fsr"       1]

    lappend format_list [list "-h" "alias --help"]
    lappend format_list [list "-v" "alias --verbose"]
    lappend format_list [list "-V" "alias --verboser"]
    lappend format_list [list "-p" "alias --precision"]
    lappend format_list [list "-P" "alias --pfile"]
    lappend format_list [list "-E" "alias --tfile"]

    set usage ""
    append usage "Usage: tabs-verify \[options\] CELL-NAME\n"
    append usage "Options:\n"
    append usage "  -h, --help                  Print this message.\n"
    append usage "  -v, --verbose               Turn on verbosity.\n"
    append usage "  -V, --verboser              Turn on space3d verbosity.\n"
    append usage "  -E, --tfile FILE.t          Use the given .t file. By default,\n"
    append usage "                              the element definition file of the\n"
    append usage "                              current project is used.\n"
    append usage "  -P, --pfile FILE.p          Use the given .p file. By default,\n"
    append usage "                              the space parameter file of the\n"
    append usage "                              current project is used.\n"
    append usage "  -p, --precision low|medium|high\n"
    append usage "                              Use this precision. It overrules the\n"
    append usage "                              following space cap3d parameters:\n"
    append usage "                              be_mode:  be_window:  max_be_area:\n"
    append usage "                         low:    0c   1*fsr*min-fs  8*min-fs^2\n"
    append usage "                      medium:    0g   2*fsr*min-fs  8*min-fs^2\n"
    append usage "                        high:    1g   5*fsr*min-fs  4*min-fs^2\n"
    append usage "  --min-fs                    Use this minimum feature size in\n"
    append usage "                              combination with option -p.\n"
    append usage "                              Default value is 1 (micron).\n"
    append usage "  --fsr                       Use this feature size ratio in\n"
    append usage "                              combination with option -p.\n"
    append usage "                              Default value is 10.\n"
    append usage "  --no-color                  Disable colored output.\n"

    std::try {
        array unset ::main_options
        std::options_parse $format_list $args ::main_options
    }  when user_exception {msg} {
        puts stderr $msg
        puts stderr $usage
        exit 1
    }

    if [info exists ::main_options(--no-color)] {
        std::disable_color
    }

    if [info exists ::main_options(--verbose)] {
        std::enable_verbosity
    } elseif [info exists ::main_options(--verboser)] {
        std::enable_verbosity
    }

    if [info exists ::main_options(--help)] {
        puts stderr $usage
        exit 0
    }

    if [info exists ::main_options(--precision)] {
        if {![string equal $::main_options(--precision) "low"] &&
            ![string equal $::main_options(--precision) "medium"] &&
            ![string equal $::main_options(--precision) "high"]} {

            std::error_message "Please specify low|medium|high as argument to --precision."
            exit 1
        }
    } else {
        set ::main_options(--precision) ""
    }

    if {[llength $::main_options(--)] != 1} {
        std::error_message "Please specify exactly one cell-name."
        puts stderr $usage
        exit 1
    }
    set cell_name [lindex $::main_options(--) 0]

    set tfile ""
    if [info exists ::main_options(--tfile)] {
        set tfile [lindex $::main_options(--tfile) 0]
	if ![file readable $tfile] {
	    std::throw user_exception "Element definition file `$tfile' does not exist or is not readable."
	}
    }

    set pfile ""
    if [info exists ::main_options(--pfile)] {
        set pfile [lindex $::main_options(--pfile) 0]
	if ![file readable $pfile] {
	    std::throw user_exception "Parameter file `$pfile' does not exist or is not readable."
	}
    }

    compute_capacitances $cell_name $tfile $pfile $::main_options(--precision)

    std::verbose "Done."
    exit 0
}


#===============================================================================
#
#  Compute capacitances.
#
#===============================================================================

proc compute_capacitances {cell_name tfile pfile precision} {

    if {[string length $precision]} {
	# feature size in micron
	set min_fs 1
	if [info exists ::main_options(--min-fs)] {
	    set min_fs $::main_options(--min-fs)
	}
	set fsratio 10
	if [info exists ::main_options(--fsr)] {
	    set fsratio $::main_options(--fsr)
	}

	if {$precision == "low"} {
	    set be_mode "0c"
	    set be_window_factor   1.0
	    set max_be_area_factor 8.0
	} elseif {$precision == "medium"} {
	    set be_mode "0g"
	    set be_window_factor   2.0
	    set max_be_area_factor 8.0
	} else {
	    set be_mode "1g"
	    set be_window_factor   5.0
	    set max_be_area_factor 4.0
	}
	set be_window   [expr $be_window_factor*$fsratio*$min_fs]
	set max_be_area [expr $max_be_area_factor*$min_fs*$min_fs]
    }

    eval {
        set cmd [list]
        lappend cmd space3d -F -3C
	if [string length $tfile] { lappend cmd -E $tfile }
	if [string length $pfile] { lappend cmd -P $pfile }
	if [string length $precision] {
	    lappend cmd -Scap3d.be_mode=$be_mode
	    lappend cmd -Scap3d.max_be_area=[format %.3g $max_be_area]
	    lappend cmd -Scap3d.be_window=[format %.3g $be_window]
	}
	if [info exists ::main_options(--verboser)] { lappend cmd -v }
        lappend cmd $cell_name

        std::verbose "Running Space in 3D mode ..."
        eval std::quiet_execute $cmd

        get_test_results $cell_name
    }

    eval {
        set cmd [list]
        lappend cmd space3d -F -Cl
	if [string length $tfile] { lappend cmd -E $tfile }
	if [string length $pfile] { lappend cmd -P $pfile }
	if [info exists ::main_options(--verboser)] { lappend cmd -v }
        lappend cmd $cell_name

        std::verbose "Running Space in heuristical mode ..."
        eval std::quiet_execute $cmd

        get_test_results $cell_name
    }
}


proc get_test_results {cell_name} {

    set circuit [circuit load $cell_name]
    set terminals [lsort [$circuit info terminals]]
    puts "terminals: $terminals"
    $circuit destroy

    foreach terminal1 $terminals {
        foreach terminal2 $terminals {

            set circuit [circuit load $cell_name]
            catch {$circuit create node 0}

            set node1 [$circuit terminal get $terminal1]
            set node2 [$circuit terminal get $terminal2]

            if {$node1 == $node2} {
                $circuit destroy
                continue
            }

            foreach gnd {gnd GND} {
                catch {$circuit collapse "0 $gnd"}
            }

            foreach terminal3 [$circuit info terminals] {
                set node3 [$circuit terminal get $terminal3]
                if {$node3 == $node1 || $node3 == $node2} continue
                $circuit collapse "0 $node3"
            }

            # Create a very small artificial conductance to ground, from each node.
            #
            foreach node [$circuit info nodes] {
                if {$node == "0"} continue
                $circuit create element res "v=1e12" 0 $node
            }

            # Create a very small artificial capacitance to ground, from each node.
            #
            foreach node [$circuit info nodes] {
                if {$node == "0"} continue
                $circuit create element cap "v=1e-21" 0 $node
            }

            set frequency 1e9
            set omega [expr 2*$::std::constants::pi*$frequency]

            set r [$circuit compute_twoport_admittances 0 $node1 $node2 $frequency]
            #puts stderr $r

            set omega_c_ground [lindex [lindex $r 0] 1]
            set c_ground [expr $omega_c_ground/$omega]

            set omega_c_bridge [lindex [lindex $r 1] 1]
            set c_bridge [expr $omega_c_bridge/$omega]
	    set cb "0"
            if {$c_bridge != 0.0} { set cb "[format %.3e $c_bridge]" }
	    puts "[format %4s $terminal1] [format %4s $terminal2]: C_gnd=[format %.3e $c_ground] C_bridge=$cb"

            $circuit destroy
        }
    }
}


#===============================================================================
#
#  Invoke main.
#
#===============================================================================

std::top_eval {
    eval main $argv
}

