# rcsid = "$Id: basic_layout.tcl,v 1.12 2014/05/07 08:34:43 simon Exp $"

std::include_once src/generic/libs/tcl/std/path.tcl
std::include_once src/generic/libs/tcl/std/format.tcl
std::include_once src/generic/libs/tcl/std/classes.tcl
std::include_once src/generic/libs/tcl/std/exception.tcl
std::include_once src/generic/libs/tcl/std/array.tcl
std::include_once src/generic/libs/tcl/std/exec.tcl
std::include_once src/generic/libs/tcl/std/math.tcl
std::include_once src/generic/libs/tcl/std/file.tcl
std::include_once src/generic/libs/tcl/std/message.tcl


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
namespace eval ::tabs::basic_layout {
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#===============================================================================
#
#  Compute capacitances.
#
#  This procedure computes coupling capacitances for a given layout, using
#  space3d. The layout should be specified in the table `layout_table'.
#  See generate_layout for a description of the layout table format.
#
#  The results are returned in the array `results'. See read_results for
#  a description of the format.
#
#===============================================================================

proc compute_capacitances {_layout_table _diel_table _precision_table _caps} {

    upvar $_layout_table layout_table
    upvar $_diel_table diel_table
    upvar $_precision_table precision_table
    upvar $_caps caps

    std::auto_chdir [::obj::get $::main_options(project) path] {

	set cell_name [::tabs::get_unique_cell_name]

	generate_layout $cell_name layout_table

	# Estimate complexity with space3d.
	#
        set cmd [list]
        lappend cmd space3d -3CU
        lappend cmd -Scap3d.be_mode=$precision_table(be_mode)
        lappend cmd -Scap3d.be_window=$precision_table(be_window)

	if {$precision_table(be_area) >= 1000} {
	    set be_area [expr round($precision_table(be_area))]
	} else {
	    set be_area [format %.3g $precision_table(be_area)]
	}
	set cmd2 $cmd
	lappend cmd2 -Scap3d.max_be_area=$be_area $cell_name
	eval exec $cmd2 >& est_dat
	set fp [open est_dat r]
	set n 23
	while {[gets $fp line] >= 0} {
	    set n [string first "elements" $line]
	    if {$n > 0} {
		set n [expr [string first ":" $line $n] + 2]
		break
	    }
	}
	close $fp
	set complexity [string range $line $n end]

	# Do a cache-lookup first.
	#
	set value ""
	set use_cache 0
	if ![info exists ::main_options(--no-cache)] { incr use_cache }
	set test_cache [info exists ::main_options(--test-cache)]
	if {$use_cache || $test_cache} {

	    # We exclude the mask names from the key.
	    # In place of that, we include space3d parameters.
	    #
	    for {set i 0} {$i < $layout_table(size)} {incr i} {
		set mask_table($i) $layout_table($i:mask)
	    }
	    set layout_table(0:mask) $precision_table(be_mode)
	    set layout_table(1:mask) $complexity
	    set layout_table(2:mask) $precision_table(be_window)

	    # Make cache key.
	    set key [list]
	    lappend key "compute_capacitances"
	    lappend key [::std::canonic_array_get layout_table]
	    lappend key [::std::canonic_array_get diel_table]

	    # Restore the mask names.
	    for {set i 0} {$i < $layout_table(size)} {incr i} {
		set layout_table($i:mask) $mask_table($i)
	    }

	    # Try to find the results in the cache.
	    set value [::tabs::cache_lookup_cap $key]
	    if {[string length $value] && !$test_cache} {
		std::verbose "Found results in cache."
		array set caps $value
		return
	    }
	}

	# Run space3d.
	#
        set cmd [list]
        lappend cmd space3d -3C

	if [info exists ::main_options(--verboser)] {lappend cmd -v}
	if [info exists ::main_options(--scene)] {lappend cmd -s "$cell_name.scn"}

        lappend cmd -Scap3d.be_mode=$precision_table(be_mode)
        lappend cmd -Scap3d.max_be_area=$be_area
        lappend cmd -Scap3d.be_window=$precision_table(be_window)
        lappend cmd $cell_name

	set fp [open "$cell_name.cmd" w]
	puts $fp "# complexity = $complexity"
	puts $fp $cmd
	close $fp

	if ![info exists ::main_options(--fake)] {

	    std::enable_verbosity
	    eval std::quiet_execute $cmd
	    std::lower_verbosity

	    read_results $cell_name layout_table caps

	    if {$test_cache} {
		if ![string length $value] {
		    puts stderr "results not found in cache"
		    incr ::main_options(nocache)
		} else {
		    array set capc $value
		    set err 0
		    foreach n [array names caps] {
			if {![info exists capc($n)] || $capc($n) != $caps($n)} {
			    incr err
			    break
			}
		    }
		    if {$err} {
			puts stderr "not equal to results in cache"
			incr ::main_options(necache)
			foreach n [lsort [array names capc]] { puts $::main_options(fp) "$n $capc($n)" }
			foreach n [lsort [array names caps]] { puts $::main_options(fp) "$n $caps($n)" }
		    } else {
			puts stderr "equal to results in cache"
			incr ::main_options(eqcache)
			set use_cache 0
		    }
		}
	    }
	    if {$use_cache} {
		::tabs::cache_insert_cap $key caps
	    }
	} else {
	    puts stderr "+ $cmd"
	}
    }
}


#===============================================================================
#
#  Generate layout (all units in 0.1 nm)
#
#  This procedure generates a new layout in the given project with the given
#  cell name.
#
#  The format of the layout table can best be described by an example:
#
#    set layout_table(0:n) 10            # number of wires (>= 1)
#    set layout_table(0:width)    10000  # width of the wire (1 um)
#    set layout_table(0:spacing)  10000  # spacing (1 um) between wires (> 0)
#    set layout_table(0:length) 1000000  # length of the wire (100 um)
#    set layout_table(0:z)         2000  # height (.2 um) of the wire (>= 0)
#    set layout_table(0:thickness)  700  # thickness (70 nm) of the wire (>= 0)
#
#    set layout_table(1:n) 10
#    set layout_table(1:width)    10000  # 1 um
#    set layout_table(1:spacing)  10000  # 1 um
#    set layout_table(1:length) 1000000  # 100 um
#    set layout_table(1:z)        12000  # 1.2 um
#    set layout_table(1:thickness)  700  # 70 nm
#
#    set layout_table(size) 2
#
#  This layout table describes two planes of wires; both planes
#  contain 10 wires.
#
#===============================================================================

proc generate_layout {cell_name _layout_table} {

    upvar $_layout_table layout_table

    set fp [open "$cell_name.ldm" w]

    puts $fp "ms $cell_name"

    for {set j 0} {$j < $layout_table(size)} {incr j} {

	if {$layout_table($j:z) == 0 && $layout_table($j:thickness) == 0} continue

	# width, length and spacing are in lambda=0.1nm units
	set w  [expr {int($layout_table($j:width))}]
	set s  [expr {int($layout_table($j:spacing))}]
	set y1 0
	if [info exists layout_table($j:y1)] { set y1 $layout_table($j:y1) }
	set y2 [expr {$y1 + int($layout_table($j:length))}]

	set n $layout_table($j:n)
	set center [expr {($n-1)/2}]

	set x1 0
	if [info exists layout_table($j:x1)] { set x1 $layout_table($j:x1) }

	for {set i 0} {$i < $n} {incr i} {
	    set x2 [expr {$x1 + $w}]
	    set name "t[encode_name $j [expr {$i-$center}]]"
	    puts $fp "term $layout_table($j:mask) $x1 $x2 $y1 $y2 $name"
	    set x1 [expr {$x2 + $s}]
	}
    }

    puts $fp "me"

    close $fp

    std::quiet_execute cldm "$cell_name.ldm"
}


#===============================================================================
#
#  Read results.
#
#  This procedure reads capacitance values from a circuit extracted using
#  space3d. The layout should have been created by generate_layout (because a
#  special node naming scheme is used).
#
#  The results are stored in the table `caps'.
#
#===============================================================================

proc read_results {cell_name _layout_table _caps} {

    upvar $_layout_table layout_table
    upvar $_caps caps

    circuit load $cell_name

    if {0 && [std::is_debug]} {
	std::debug_message "Circuit, as extracted:"
	puts stderr [$cell_name xsls]
    }

    array unset caps
    foreach line [$cell_name xsls] {
	if {[llength $line] > 3} {
	    set incap 0
	    foreach m $line {
		if {$incap} {
		    if {$incap == 1} {
			if {[regexp {p$} $m]} {
			    set l [string length $m]
			    set v [string range $m 0 [expr {$l - 2}]]e-12
			} elseif {[regexp {f$} $m]} {
			    set l [string length $m]
			    set v [string range $m 0 [expr {$l - 2}]]e-15
			} else {
			    set v $m
			}
			incr incap
		    } elseif {$incap == 2} {
			regsub -all {[(,]} $m "" n1
			incr incap
		    } else {
			regsub -all {[);]} $m "" n2
			set incap 0
			if {$n1 < $n2} {set caps($n1,$n2) $v} else {set caps($n2,$n1) $v}
		    }
		} elseif {$m == "cap"} {incr incap}
	    }
	}
    }

    if {0 && [std::is_debug]} {
        foreach n [lsort [array names caps]] {
            std::debug_message "cap($n) = $caps($n)"
        }
    }
}


#===============================================================================
#
#  Decode a name.
#
#  For example: "p23n12" is decoded into 23 and -12.
#
#===============================================================================

proc decode_name {name _value0 _value1} {

    upvar $_value0 value0
    upvar $_value1 value1

    if ![decode_number name value0] {
        return 0
    }
    if ![decode_number name value1] {
        return 0
    }
    return 1
}

proc decode_number {_string _value} {

    upvar $_string string
    upvar $_value value

    set char [string index $string 0]
    if {$char == "c"} {
        set string [string range $string 1 end]
        set value 0
        return 1
    }

    if {$char != "p" && $char != "n"} {
        return 0
    }
    set string [string range $string 1 end]

    if ![regexp -- {^[0-9]+} $string value] {
        return 0
    }

    set string [string range $string [string length $value] end]

    regsub {^[0]+} $value "" value
    if ![string length $value] {set value 0}
    if {$char == "n"} {set value [expr -int($value)]}
    return 1
}


#===============================================================================
#
#  Encode a name.
#
#  For example: 23 and -12 are encoded into p23n12.
#
#===============================================================================

proc encode_name {index0 index1} {
    return "[encode_number $index0][encode_number $index1]"
}

proc encode_number {index} {

    if {$index < 0} {
        return "n[expr -int($index)]"
    }
    if {$index > 0} {
        return "p[expr int($index)]"
    }
    return "c"
}


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
}
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


