# rcsid = "$Id: tools.tcl,v 1.10 2011/11/28 08:24:38 simon Exp $"

std::include_once src/generic/libs/tcl/std/path.tcl
std::include_once src/generic/libs/tcl/std/format.tcl
std::include_once src/generic/libs/tcl/std/classes.tcl
std::include_once src/generic/libs/tcl/std/exception.tcl
std::include_once src/generic/libs/tcl/std/array.tcl
std::include_once src/generic/libs/tcl/std/exec.tcl
std::include_once src/generic/libs/tcl/std/math.tcl
std::include_once src/generic/libs/tcl/std/file.tcl
std::include_once src/generic/libs/tcl/std/message.tcl
std::include_once src/generic/libs/tcl/std/classes.tcl

std::include_once src/space/utilities/libs/tcl/project_tools/read_technology.tcl


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
namespace eval tabs {
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

variable temporary_directory
catch {unset temporary_directory}

variable cache_directory
catch {unset cache_directory}

variable default_lambda
catch {unset default_lambda}

variable __is_initialized
set __is_initialized 0

proc initialize {} {

    std::invariant !$::tabs::__is_initialized

    set ::tabs::temporary_directory [std::tmppath "tabs"]
    if ![file exists $::tabs::temporary_directory] {
        ::std::verbose "Creating project directory $::tabs::temporary_directory ..."
        file mkdir $::tabs::temporary_directory
    }

    foreach dir {tabs tabs2 tabs3} {
	set old $::env(HOME)/.cacd/cache/$dir
	if [file exists $old] {
	    std::warning_message "Old cache directory `$old' found (can be removed)."
	}
    }

    set ::tabs::cache_directory $::env(HOME)/.cacd/cache/tabs4
    if ![file exists $::tabs::cache_directory] {
        file mkdir $::tabs::cache_directory
    }

    set ::tabs::default_lambda 0.0001

    set ::tabs::__is_initialized 1
}


#===============================================================================
#
#  Cell names.
#
#===============================================================================

variable last_cell_id
set last_cell_id 0

proc get_unique_cell_name {} {
    std::invariant $::tabs::__is_initialized
    return [format cell%04d [incr ::tabs::last_cell_id]]
}


#===============================================================================
#
#  Create a project.
#
#===============================================================================

set ::last_project_id 0

::obj::template {::tabs::project} {

    ::obj::inherit ::object
    ::obj::attribute path %notype
    ::obj::attribute lambda %notype
}

proc create_project {process_path} {

    std::invariant $::tabs::__is_initialized

  # set prefix "tabs"
  # while 1 {
  #     set project_name "${prefix}-[format %04d [incr ::last_project_id]]"
  #     set project_path $::tabs::temporary_directory/${project_name}
  #     if ![file exists $project_path] break
  # }

    set project_path $::tabs::temporary_directory

if ![file exists $project_path/.dmrc] {
    ::std::verbose "Making project directory $project_path ..."
    ::std::quiet_execute mkpr -p $process_path -l $::tabs::default_lambda $project_path
}

    set project [::obj::new ::tabs::project]
    ::obj::assign $project path $project_path
    ::obj::assign $project lambda $::tabs::default_lambda

    return $project
}

proc get_tecc_color {i} {

    set color_list {1 4 3 2 5 6 7}
    set color [lindex $color_list [expr $i%[llength $color_list]]]

    set red   [format %02x [expr (($color>>0)&1)*255]]
    set green [format %02x [expr (($color>>1)&1)*255]]
    set blue  [format %02x [expr (($color>>2)&1)*255]]

    return "@$red$green$blue"
}


#===============================================================================
#
#  Ideal plate capacitance.
#
#===============================================================================

proc get_ideal_plate_capacitance_between_layers {h1 h2 _technology_table} {

    upvar $_technology_table technology_table

    std::invariant $::tabs::__is_initialized

    # Use relative permittivity of silicon as start value.
    set previous_permittivity 11.7
    set previous_z $h1
    set list [list]

    for {set j 0} {$j < $technology_table(dielectrics:size)} {incr j} {

        set z $technology_table(dielectrics:$j:z)
        set permittivity $technology_table(dielectrics:$j:permittivity)

        if {$z >= $h1 && $z < $h2} {
            if {$z > $previous_z} {
                lappend list [list [expr $z - $previous_z] $previous_permittivity]
                set previous_z $z
            }
        } elseif {$z >= $h2} {
            break
        }

        set previous_permittivity $permittivity
    }

    lappend list [list [expr $h2 - $previous_z] $previous_permittivity]

    set tmp 0.0

    foreach tuple $list {
        set h  [expr [lindex $tuple 0]/1e10]
        set er [lindex $tuple 1]
        set cpp [expr $er*$::std::constants::epsilon_0/$h]
        set tmp [expr $tmp + 1/$cpp]
    }

    return [expr 1/$tmp]
}

proc edge_condition {condition {type "-"}} {

    set condition [libmin::minimize $condition]

    if {$condition == "0"} {return "0"}
    if {$condition == "1"} {return "1"}

    set list [list]
    foreach w [split $condition " "] {
        if {$w == "|"} {
            lappend list $w
            continue
        }

        set prefix ""
        if {[string index $w 0] == "!"} {
            set prefix "!"
            set w [string range $w 1 end]
        }

        if {[string index $w 0] == "-" || [string index $w 0] == "="} {
            # This case should be an error, actually.
            lappend list "${prefix}${w}"
            continue
        }

        lappend list "${prefix}${type}${w}"
    }

    return [join $list " "]
}


#===============================================================================
#
#  Cache.
#
#===============================================================================

proc cache_lookup {key} {

    set hash [digest $key]
    set path $::tabs::cache_directory/${hash}.cache
    if ![file readable $path] {
        return ""
    }

    set fp [open $path r]
    gets $fp key_line
    gets $fp value_line
    close $fp

    set found_key [base64 decode $key_line]
    set found_value [base64 decode $value_line]

    if ![string equal $key $found_key] {
        return ""
    }

    return $found_value
}

proc cache_insert {key value} {

    set hash [digest $key]
    set path $::tabs::cache_directory/${hash}.cache

    set fp [open $path w]
    puts $fp [base64 encode $key]
    puts $fp [base64 encode $value]
    close $fp
}

proc cache_insert_cap {key _caps} {

    upvar $_caps cap
    set fp [open "$::tabs::cache_directory/[digest $key].cap" w]
    puts $fp [array get cap]
    close $fp
}

proc cache_lookup_cap {key} {

    set path "$::tabs::cache_directory/[digest $key].cap"
    set value ""
    if [file readable $path] {
	set fp [open "$path" r]
	gets $fp value
	close $fp
    }
    return $value
}


#===============================================================================
#
#  Progress computation.
#
#===============================================================================

set ::progress_enabled 0
set ::progress_priming 0
set ::progress 0
set ::progress_total 0

proc progress_enable {} {
    set ::progress_enabled 1
}

proc progress_prime {script} {
    if !$::progress_enabled return
    set ::progress_priming 1

    uplevel #1 $script

    set ::progress_priming 0
    set ::progress_total $::progress
    set ::progress 0
}

proc progress_add {} {
    if !$::progress_enabled { return 0 }
    incr ::progress

    if !$::progress_priming {
        set msg "$::progress/$::progress_total"
        puts stderr "progress: [std::colorize $msg blue]"
    }
    return $::progress_priming
}

proc progress_finish {} {
    if !$::progress_enabled return

    std::invariant !$::progress_priming

    if {$::progress < $::progress_total} {
        set msg "$::progress_total/$::progress_total"
        puts stderr "progress: [std::colorize $msg blue]"
    }
}


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
}
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

